//
// view.c
// -OSX audio input visualization by Stanko Juzbasic
// -based on and including parts of 'AuhalGlut' code by:
//   Dmitrij Hmeljak -- http://www.cs.indiana.edu/~mitja/
// -also contains parts of F. Richard Moore code from
//      Elements of Computer Music
//
//cc -o view view.c -framework Accelerate AudioToolbox AudioUnit CoreAudio CoreServices GLUT OpenGL


#include <stdio.h>
#include <string.h>
#include <CoreServices/CoreServices.h> 
#include <AudioToolbox/AudioToolbox.h>
#include <AudioUnit/AudioUnit.h> 
#include <Accelerate/Accelerate.h>



#include <GLUT/glut.h> // for the graphic demo



// ------------------------------------------------------------------------
// audio constants
#define kOutputBus 0
#define kInputBus  1
// constants for GLUT window defaults & math
#define MY02TEST_WIND_X_SIZE        640
#define MY02TEST_WIND_Y_SIZE        640
#define MY02TEST_WIND_X_MINSIZE     128
#define MY02TEST_WIND_Y_MINSIZE     128
#define MY02TEST_WIND_X_POS         100
#define MY02TEST_WIND_Y_POS         100

#ifndef PI
#define PI       3.141592653589793238462643
#define TWOPI    6.283185307179586476925287
#endif

#define tN 2048

//____________________________________________________________________________
// macros
#ifdef MAX
#undef MAX
#endif
#ifndef MAX
#define MAX(A,B) ((A) >= (B) ? (A) : (B))
#endif

#ifdef MIN
#undef MIN
#endif
#ifndef MIN
#define MIN(A,B) ((A) <= (B) ? (A) : (B))
#endif

//____________________________________________________________________________

// global variables

char * version = "0.9j";

AudioDeviceID inputDevice;

int gLastKey = ' ';

Boolean       poly; //polygon drawing mode

Boolean       dB;

Boolean       logfrq;

Boolean       windowing;

Boolean       sonogram;

Boolean       surface;

Float64       R;

UInt32        D;
UInt32        N;
UInt32        N2;

unsigned int D1;
unsigned int DN, D1N;

float minvalue = 0.00001;

static int           AUHAL_SAMPLES = 512;

static int           gNpointsToDisplay;

static int           bgNpointsToDisplay;

// the Audio Unit connected to Hardware Abstraction Layer's input device:
static ComponentInstance gInputUnitComponentInstance;

// buffers used during AUHAL rendering *only* :
static AudioBufferList *gAudioBufferList;
// buffers used elsewhere:
static Float32 gvAudioBuffers[2][tN];

// input audio characteristics:
static AudioStreamBasicDescription gAudioStreamDescription;

float gn = 1.;//display gain

float lgn= 1.;//logarithmic display gain

int   vt = 0.;//vertical translator;


DSPSplitComplex mComplexData [tN];
float realpart [tN];
float imagpart [tN];
Float32 spectrumData [tN];
Float32 ampData [tN];
Float32 frqData [tN];
float fundamental;
float lastphase [tN];
float *win;
FFTSetup  mFFTSetup;

DSPSplitComplex bmComplexData [8*tN];
float brealpart [8*tN];
float bimagpart [8*tN];
Float32 bspectrumData [8*tN];
Float32 bampData [8*tN];
Float32 bfrqData [8*tN];
float bfundamental;
float blastphase [8*tN];
float *bwin;
FFTSetup bmFFTSetup;

Float32 *megamp, *megfreq;
Float32 *megamphead, *megfreqhead;
unsigned int count;


int gY;
Boolean debug = false;
Boolean b = false;
Boolean freqDisp = false;


//_________________________________________________________________________________
// function prototypes
void cheby_win(float *out, int N, float atten);
double cheby_poly(int n, double x);
void  GlSpecial (int, int, int);
char* descriptionForFormatFlags(UInt32  mFormatFlags);

static void GlBitmapString(int x, int y, char *string);

static void GlOutputStart(void);

static void GlInit(void);

static void GlDisplay(void);
static void reshape(int width, int height);
static void GlKey(unsigned char key, int x, int y);
static void GlIdle(void);
//__________________________________________________________________________________
/***************************************************************************
 calculate a chebyshev window of size N, store coeffs in out as in Antoniou
 -out should be array of size N
 -atten is the required sidelobe attenuation (e.g. if you want -60dB atten, use '60')
 ***************************************************************************/
void cheby_win(float *out, int N, float atten){
    int nn, i;
    double M, n, sum = 0, max=0;
    double tg = pow(10,atten/20);  /* 1/r term [2], 10^gamma [2] */
    double x0 = cosh((1.0/(N-1))*acosh(tg));
    M = (N-1)/2;
    if(N%2==0) M = M + 0.5; /* handle even length windows */
    for(nn=0; nn<(N/2+1); nn++){
        n = nn-M;
        sum = 0;
        for(i=1; i<=M; i++){
            sum += cheby_poly(N-1,x0*cos(PI*i/N))*cos(2.0*n*PI*i/N);
        }
        out[nn] = tg + 2*sum;
        out[N-nn-1] = out[nn];
        if(out[nn]>max)max=out[nn];
    }
    for(nn=0; nn<N; nn++) out[nn] /= max; /* normalise everything */
    return;
}

//**************************************************************************
//  This function computes the chebyshev polyomial T_n(x)
//***************************************************************************
double cheby_poly(int n, double x){
    double res;
    if (fabs(x) <= 1) res =  cos(n*acos (x));
    else              res = cosh(n*acosh(x));
    return res;
}

//__________________________________amp2db______________________________________________
float amp2dbf(float avalue, float minvalue)	//converts amplitude ratios to db
{
    
    if( avalue <= minvalue ) return 0.;
    else avalue = 20.*log10f(avalue/minvalue);
    return avalue;
}
//______________________________________________________________________________________
char* descriptionForFormatFlags(UInt32  mFormatFlags)
{
    //NSMutableString *description = [NSMutableString new];
    static char description[512];
    
    if (mFormatFlags & kAudioFormatFlagIsFloat)
    { //[description appendString:@"kAudioFormatFlagIsFloat \n"];
        strcat(description, "kAudioFormatFlagIsFloat \n");}
    if (mFormatFlags & kAudioFormatFlagIsBigEndian)
    { //[description appendString:@"kAudioFormatFlagIsBigEndian \n"];
        strcat(description, "kAudioFormatFlagIsBigEndian \n");}
    if (mFormatFlags & kAudioFormatFlagIsSignedInteger)
    { //[description appendString:@"kAudioFormatFlagIsSignedInteger \n"];
        strcat(description, "kAudioFormatFlagIsSignedInteger \n");}
    if (mFormatFlags & kAudioFormatFlagIsPacked)
    { //[description appendString:@"kAudioFormatFlagIsPacked \n"];
       strcat(description, "kAudioFormatFlagIsPacked \n");}
    if (mFormatFlags & kAudioFormatFlagIsAlignedHigh)
    { //[description appendString:@"kAudioFormatFlagIsAlignedHigh \n"];
       strcat(description, "kAudioFormatFlagIsAlignedHigh \n");}
    if (mFormatFlags & kAudioFormatFlagIsNonInterleaved)
    { //[description appendString:@"kAudioFormatFlagIsNonInterleaved \n"];
       strcat(description, "kAudioFormatFlagIsNonInterleaved \n");}
    if (mFormatFlags & kAudioFormatFlagIsNonMixable)
    { //[description appendString:@"kAudioFormatFlagIsNonMixable \n"];
       strcat(description, "kAudioFormatFlagIsNonMixable \n");}
    if (mFormatFlags & kAudioFormatFlagsAreAllClear)
    { //[description appendString:@"kAudioFormatFlagsAreAllClear \n"];
      strcat(description, "kAudioFormatFlagsAreAllClear \n");}
    if (mFormatFlags & kLinearPCMFormatFlagIsFloat)
    { //[description appendString:@"kLinearPCMFormatFlagIsFloat \n"];
      strcat(description, "kLinearPCMFormatFlagIsFloat \n");}
    if (mFormatFlags & kLinearPCMFormatFlagIsBigEndian)
    { //[description appendString:@"kLinearPCMFormatFlagIsBigEndian \n"];
        strcat(description, "kLinearPCMFormatFlagIsBigEndian \n");}
    if (mFormatFlags & kLinearPCMFormatFlagIsSignedInteger)
    { //[description appendString:@"kLinearPCMFormatFlagIsSignedInteger \n"];
        strcat(description, "kLinearPCMFormatFlagIsSignedInteger \n");}
    if (mFormatFlags & kLinearPCMFormatFlagIsPacked)
    { //[description appendString:@"kLinearPCMFormatFlagIsPacked \n"];
        strcat(description, "kLinearPCMFormatFlagIsPacked \n");}
    if (mFormatFlags & kLinearPCMFormatFlagIsAlignedHigh)
    { //[description appendString:@"kLinearPCMFormatFlagIsAlignedHigh \n"];
        strcat(description, "kLinearPCMFormatFlagIsAlignedHigh \n");}
    if (mFormatFlags & kLinearPCMFormatFlagIsNonInterleaved)
    { //[description appendString:@"kLinearPCMFormatFlagIsNonInterleaved \n"];
        strcat(description, "kLinearPCMFormatFlagIsNonInterleaved \n");}
    if (mFormatFlags & kLinearPCMFormatFlagIsNonMixable)
    { //[description appendString:@"kLinearPCMFormatFlagIsNonMixable \n"];
        strcat(description, "kLinearPCMFormatFlagIsNonMixable \n");}
    if (mFormatFlags & kLinearPCMFormatFlagsSampleFractionShift)
    { //[description appendString:@"kLinearPCMFormatFlagsSampleFractionShift \n"];
        strcat(description, "kLinearPCMFormatFlagsSampleFractionShift \n");}
    if (mFormatFlags & kLinearPCMFormatFlagsSampleFractionMask)
    { //[description appendString:@"kLinearPCMFormatFlagsSampleFractionMask \n"];
        strcat(description, "kLinearPCMFormatFlagsSampleFractionMask \n");}
    if (mFormatFlags & kLinearPCMFormatFlagsAreAllClear)
    { //[description appendString:@"kLinearPCMFormatFlagsAreAllClear \n"];
        strcat(description, "kLinearPCMFormatFlagsAreAllClear \n");}
    if (mFormatFlags & kAppleLosslessFormatFlag_16BitSourceData)
    { //[description appendString:@"kAppleLosslessFormatFlag_16BitSourceData \n"];
        strcat(description, "kAppleLosslessFormatFlag_16BitSourceData \n");}
    if (mFormatFlags & kAppleLosslessFormatFlag_20BitSourceData)
    { //[description appendString:@"kAppleLosslessFormatFlag_20BitSourceData \n"];
        strcat(description, "kAppleLosslessFormatFlag_20BitSourceData \n");}
    if (mFormatFlags & kAppleLosslessFormatFlag_24BitSourceData)
    {  //[description appendString:@"kAppleLosslessFormatFlag_24BitSourceData \n"];
        strcat(description, "kAppleLosslessFormatFlag_24BitSourceData \n");}
    if (mFormatFlags & kAppleLosslessFormatFlag_32BitSourceData)
    {  //[description appendString:@"kAppleLosslessFormatFlag_32BitSourceData \n"];
        strcat(description, "kAppleLosslessFormatFlag_32BitSourceData \n");}
    
    //return [NSString stringWithString:description];
    return description;
}

//________________________________________________________________
static OSStatus GetIOBufferFrameSizeRange(AudioUnit inAUHAL,
                                          UInt32* outMinimum,
                                          UInt32* outMaximum)
{
    AudioValueRange theRange = { 0, 0 };
    UInt32 theDataSize = sizeof(AudioValueRange);
    OSStatus theError = AudioUnitGetProperty(inAUHAL,
                                             kAudioDevicePropertyBufferFrameSizeRange,
                                             kAudioUnitScope_Global,
                                             0,
                                             &theRange, &theDataSize);
    
    if(theError == 0)
    {
        *outMinimum = theRange.mMinimum;
        *outMaximum = theRange.mMaximum;
    }
    return theError;
}

# pragma mark - callback -
// ----------------------------------------------------------------
// The AUHAL is an Audio Unit that can receive and send audio data to an audio device.
// To receive audio from the AUHAL, you must get it from the output scope of the Audio Unit.
// In practice, this is done by a client calling AudioUnitRender.
// ----------------------------------------------------------------
static OSStatus /* (*AURenderCallback) */ vAURenderCallback (
                                                void *inRefCon,
                                                AudioUnitRenderActionFlags *ioActionFlags,
                                                const AudioTimeStamp *inTimeStamp,
                                                UInt32 inBusNumber,
                                                UInt32 inNumberFrames,
                                                AudioBufferList *ioData )
                                                //AudioBufferList *gAudioBufferList)
{
    OSStatus lErr = noErr;
    int i;
    Float32 * lBuffer0;
    Float32 * lBuffer1;
    Float32 buf [inNumberFrames];
    Float32 phase, phasediff;
    Float32 factor = (Float32)R/(inNumberFrames*TWOPI);
    static Boolean firstErr;
    

    
    mComplexData->realp = realpart;
    mComplexData->imagp = imagpart;
    
    Float32 bbuf[N*2];
    Float32 bigbuf[N*2];
    
    bmComplexData->realp = brealpart;
    bmComplexData->imagp = bimagpart;
    
    gNpointsToDisplay = inNumberFrames;
    bgNpointsToDisplay = N;
   
    
    if(0){
    static Float64 prev_stime;  //prev. sample time
    static UInt64  prev_htime;  //prev. host time     
    printf("inBus: %d\tframes: %d\tHtime: %lld\tStime: %7.2lf\n",
           (unsigned int)inBusNumber,
           (unsigned int)inNumberFrames,
           inTimeStamp->mHostTime   - prev_htime,
           inTimeStamp->mSampleTime - prev_stime);
    prev_htime = inTimeStamp->mHostTime;
    prev_stime = inTimeStamp->mSampleTime;
    }
    
    // You must call the AUHAL's AudioUnitRender from within the input callback function.
    // The input callback's render action flags, time stamp, bus number
    //    and number of frames requested
    //    should be propagated down to the AudioUnitRender call.
    lErr= AudioUnitRender(gInputUnitComponentInstance,
                         ioActionFlags,
                         inTimeStamp,
                         inBusNumber,     //will be '1' for input data
                         //kInputBus,
                         inNumberFrames,  //# of frames requested
                         gAudioBufferList);
 //                        ioData);


    lBuffer0 = (Float32 *)gAudioBufferList->mBuffers[0].mData;
    lBuffer1 = (Float32 *)gAudioBufferList->mBuffers[1].mData;

    
    
    if (lErr){
        if(!firstErr){
        if(lErr == -10863){
            printf("Couldn't render in current context\n\
\tPlease change the input sample rate to a \n\
\tmultiple of 44100 Hz (Error -10863)\n");
        }else if(lErr == -10851){
            printf("kAudioUnitErr_InvalidPropertyValue (Error -10851)\n");
        }else printf("Render Error: %d!\n",(int)lErr);
        firstErr = true;
        }
    } else {
     //for inNumberFrames analysis
        for (i=0;i<inNumberFrames;i++) {
            gvAudioBuffers[0][i]=lBuffer0[i];
            gvAudioBuffers[1][i]=lBuffer0[i];
            if(windowing)
            buf[i] = lBuffer0[i]*win[i];
            else
            buf[i] = lBuffer0[i];
        } //
        
 
    
    vDSP_ctoz((DSPComplex *)buf, 2, mComplexData, 1, inNumberFrames);
    //FFTSetup mFFTSetup = vDSP_create_fftsetup ( 9, 2 );
    vDSP_fft_zrip(mFFTSetup, mComplexData, 1, log2f(inNumberFrames), FFT_FORWARD);
    mComplexData->imagp[0] = 0.0;
    // Scale the data
    float scale = (float) 1.0 / (2 * (float)inNumberFrames);
    vDSP_vsmul(mComplexData->realp, 1, &scale, mComplexData->realp, 1, inNumberFrames);
    vDSP_vsmul(mComplexData->imagp, 1, &scale, mComplexData->imagp, 1, inNumberFrames);
    
    // Convert the complex data into magnitude...
    // spectrumData is also a (float*) of size mNumFrequencies
    vDSP_zvabs(mComplexData, 1, ampData, 1, inNumberFrames);
    
    // ...and frequency
    for (int j = 0; j < inNumberFrames; j++ ){
        Float32 a = mComplexData->realp[j];
        Float32 b = mComplexData->imagp[j];
        if(ampData[j] == 0.) phasediff = 0.;
        else{
            phasediff = ( phase = -atan2( b, a ) ) - lastphase[j];            
            //unwrap
            while ( phasediff > PI )
                phasediff -= TWOPI;
            while ( phasediff < -PI )
                phasediff += TWOPI;
        }
        frqData [j] = phasediff*factor + j*fundamental;
    }// end j
        
        
     //for full FFT size analysis
         //shift-in+folding;
         for(int j = D; j < N; j++)
         bigbuf[j - D] = bigbuf[j];
         for(i = 0; i < D; i++)
         bigbuf[i + N - D] = lBuffer0[i];
         memcpy(bbuf,bigbuf,N*sizeof(Float32));
         //windowing:
         for (int j = 0; j < N; j++){
         bbuf[j] *= bwin[j];
         }
         
         
        vDSP_ctoz((DSPComplex *)bbuf, 2, bmComplexData, 1, N);
        vDSP_fft_zrip(bmFFTSetup, bmComplexData, 1, log2f(N), FFT_FORWARD);
        mComplexData->imagp[0] = 0.0;
        // Scale the data
        float bscale = (float) 1.0 / (2 * (float)N);
        vDSP_vsmul(bmComplexData->realp, 1, &bscale, bmComplexData->realp, 1, N);
        vDSP_vsmul(bmComplexData->imagp, 1, &bscale, bmComplexData->imagp, 1, N);
        
        // Convert the complex data into something usable
        // ampData is also a (float*) of size mNumFrequencies
        vDSP_zvabs(bmComplexData, 1, bampData, 1, N);
        
        for (int bj = 0; bj < N; bj++ ){
            Float32 a = bmComplexData->realp[bj];
            Float32 b = bmComplexData->imagp[bj];
            if(bampData[bj] == 0.) phasediff = 0.;
            else{
                phasediff = ( phase = -atan2( b, a ) ) - lastphase[bj];
                
                //unwrap
                while ( phasediff > PI )
                    phasediff -= TWOPI;
                while ( phasediff < -PI )
                    phasediff += TWOPI;
            }
            bfrqData [bj] = phasediff*factor + bj*fundamental;    
        }// end bj    
   } 
    
        
    return lErr;
    
} // vAURenderCallback()


// ----------------------------------------------------------------
OSStatus vAUOutputOpen_01() {
// ----------------------------------------------------------------
    // Obtain an AudioOutputUnit by using a Component Description, as for any other Audio Unit.
    OSErr lErr;
    Component lComponent;
    ComponentDescription lComponentDescription;
    
    fprintf(stdout,"------------- (1) Open an AudioOutputUnit  (1) -------------\n");
    //There are several different types of Audio Units.
    //Some audio units serve as Outputs, Mixers, or DSP units. See AUComponent.h for listing
    lComponentDescription.componentType = kAudioUnitType_Output;
    
    //Every Component has a subType, which will give a clearer picture
    //of what this components function will be.
    lComponentDescription.componentSubType = kAudioUnitSubType_HALOutput;
    
     //all Audio Units in AUComponent.h must use
     //"kAudioUnitManufacturer_Apple" as the Manufacturer
    lComponentDescription.componentManufacturer = kAudioUnitManufacturer_Apple;
    lComponentDescription.componentFlags = 0; // - TODO - what do these flags/masks mean????
    lComponentDescription.componentFlagsMask = 0;
    
    //Finds a component that meets the lComponentDescription spec's:
    lComponent = FindNextComponent(NULL, &lComponentDescription);
    if (lComponent == NULL) exit (-1);
    
    // gains access to the services provided by the component:
    lErr = OpenAComponent(lComponent, &gInputUnitComponentInstance);
     // interesting side-note: OpenAComponent is:
     // available in CarbonLib 1.0 and later when QuickTime 2.5 (?!) or later is present.
     // Available in Mac OS X 10.0 and later.
     // Available in QuickTime for Windows. Brrrrrr...
    return lErr;
} // vAUOutputOpen_01()


// ----------------------------------------------------------------
OSStatus vAUHALEnableInputIO_02() {
// ----------------------------------------------------------------
    OSStatus lErr =noErr;

    fprintf(stdout,"------------- (2) Enable HAL IO for input  (2) -------------\n");
    // After creating the AUHAL object,
    // you must enable IO on the input scope of the Audio Unit to obtain device input.
    // Input must be explicitly enabled with the kAudioOutputUnitProperty_EnableIO property
    // on element 1 of the AUHAL.
    UInt32 lEnableIO;
//    UInt32 lSize=0;
    
     //When using AudioUnitSetProperty the 4th parameter in the method
     //refer to an AudioUnitElement.  When using an AudioOutputUnit
     //the input element will be '1' and the output element will be '0'.
    lEnableIO = 1;
    lErr = AudioUnitSetProperty(gInputUnitComponentInstance,
                         kAudioOutputUnitProperty_EnableIO,
                         kAudioUnitScope_Input,
                         kInputBus, // input element
                         &lEnableIO,
                         sizeof(lEnableIO));
    if (lErr) return lErr;

    // Because the AUHAL can be used for both input and output,
    // for this example we must also disable IO on the output scope.
    lEnableIO = 0;
    lErr = AudioUnitSetProperty(gInputUnitComponentInstance,
                         kAudioOutputUnitProperty_EnableIO,
                         kAudioUnitScope_Output,
                         kOutputBus,   //output element
                         &lEnableIO,
                         sizeof(lEnableIO));
    return lErr;
} // vAUHALEnableInputIO_02()


// ----------------------------------------------------------------
OSStatus vAUHALSetDefaultInputDev_03() {
// ----------------------------------------------------------------
    fprintf(stdout,"------------- (3) Set default In : current (3) -------------\n");
    UInt32 lSize = sizeof(AudioDeviceID);
    OSStatus lErr =noErr;
    
    // AudioHardwareGetProperty used with the parameter
    // kAudioHardwarePropertyDefaultInputDevice will obtain
    // the current input device selected by the user. 
    //AudioDeviceID inputDevice;
    lErr = AudioHardwareGetProperty(kAudioHardwarePropertyDefaultInputDevice,
                                   &lSize,
                                   &inputDevice);
    if (lErr) return lErr;
    lErr =AudioUnitSetProperty(gInputUnitComponentInstance,
                              kAudioOutputUnitProperty_CurrentDevice,
                              kAudioUnitScope_Global,
                              kOutputBus,
                              &inputDevice,
                              sizeof(inputDevice));
    
    UInt32 propertySize = sizeof(inputDevice);
    AudioObjectPropertyAddress inputDeviceProperty;
    //inputDeviceProperty.mSelector = kAudioHardwarePropertyDefaultInputDevice;
    inputDeviceProperty.mSelector = kAudioDevicePropertyAvailableNominalSampleRates;
    inputDeviceProperty.mScope    = kAudioObjectPropertyScopeGlobal;
    inputDeviceProperty.mElement  = kAudioObjectPropertyElementMaster;
    
    lErr = AudioObjectGetPropertyDataSize(inputDevice,
                                        &inputDeviceProperty,
                                        0,
                                        NULL,
                                        &propertySize);
    
    if (lErr != noErr) printf("Couldn't get sample rate count");
    
    int m_valueCount = propertySize / sizeof(AudioValueRange) ;
    
    printf("Available %d Sample Rates\n",m_valueCount) ;
    
    AudioValueRange m_valueTabe[m_valueCount];
    
    lErr = AudioObjectGetPropertyData(inputDevice,
                                        &inputDeviceProperty,
                                        0,
                                        NULL,
                                        &propertySize,
                                        m_valueTabe);
    if (lErr != noErr) printf("Couldn't get available sample rates");
    
    for(UInt32 i = 0 ; i < m_valueCount ; ++i)
    {
        printf("Available Sample Rate value : %f\n", m_valueTabe[i].mMinimum) ;
    }
    // Set the sample rate to one of the available values.
    /*  */
    AudioValueRange inputSampleRate;
    inputSampleRate.mMinimum = 44100;
    inputSampleRate.mMaximum = 48000;
    inputDeviceProperty.mSelector = kAudioDevicePropertyNominalSampleRate;
    lErr = AudioObjectSetPropertyData(inputDevice,
                                    &inputDeviceProperty,
                                    0,
                                    NULL,
                                    sizeof(inputSampleRate),
                                    &inputSampleRate);
    if (lErr != noErr) printf("Couldn't SET sample rate");
  
    
    return lErr;
} // vAUHALSetDefaultInputDev_03()
//_______________________________________________________________________________________
OSStatus vGetIOBufferrange_03_1(){
    OSStatus lErr =noErr;
    UInt32 minIObufSize, maxIObufSize;
    lErr = GetIOBufferFrameSizeRange(gInputUnitComponentInstance,
                                     &minIObufSize,
                                     &maxIObufSize);
    if(lErr == noErr){
    fprintf(stdout,"------------- (3_1) Get buff. size range (3_1) -------------\n");
    fprintf(stdout,"    between: %d and %d frames\n", (int)minIObufSize , (int)maxIObufSize);
        
    }
return lErr;
}

// ----------------------------------------------------------------
OSStatus vAudioOutputUnitCurrentDeviceSet_04() {

    OSStatus lErr =noErr;
    UInt32 lSize;
    
   // Setting up the desired 'input' format

    // Get the available sample rates of the default input device.
    //defaultDeviceProperty.mSelector = kAudioDevicePropertyAvailableNominalSampleRates;
    
    AudioStreamBasicDescription DeviceFormat  = { 0.0, 0, 0, 0, 0, 0, 0, 0, 0 };
    AudioStreamBasicDescription DesiredFormat = { 0.0, 0, 0, 0, 0, 0, 0, 0, 0 };
    
    fprintf(stdout,"------------- (4) Set curr. dev. Output AU (4) -------------\n");
    lSize = sizeof(AudioStreamBasicDescription);
    
    //Get the input device format
    lErr = AudioUnitGetProperty (gInputUnitComponentInstance,
                          kAudioUnitProperty_StreamFormat,
                          kAudioUnitScope_Input,
                          kInputBus,
                          &DeviceFormat,
                          &lSize);
    if (lErr) return lErr;
    /*
    for (int i = 0; i < asbdCount; i++) {
        UInt32 format4cc = CFSwapInt32HostToBig(asbds[i].mFormatID);
        Printf("mFormatID: %4.4s", (char*)&format4cc);
    }
    */
    //set the desired format to the device's sample rate
    R = DesiredFormat.mSampleRate             = DeviceFormat.mSampleRate;
    //R = DesiredFormat.mSampleRate = 44100;
    // save the necessary info about audio input signal:
    gAudioStreamDescription.mSampleRate       = DesiredFormat.mSampleRate;
    gAudioStreamDescription.mFormatID         = DeviceFormat.mFormatID;
    gAudioStreamDescription.mBytesPerPacket   = DeviceFormat.mBytesPerPacket;
    gAudioStreamDescription.mFramesPerPacket  = DeviceFormat.mFramesPerPacket;
    gAudioStreamDescription.mBytesPerFrame    = DeviceFormat.mBytesPerFrame;
    gAudioStreamDescription.mChannelsPerFrame = DeviceFormat.mChannelsPerFrame;
    gAudioStreamDescription.mBytesPerFrame    = DeviceFormat.mBytesPerFrame;
    gAudioStreamDescription.mBitsPerChannel   = DeviceFormat.mBitsPerChannel;
    gAudioStreamDescription.mFormatFlags      = DeviceFormat.mFormatFlags;
    
    // printout some information:

    fprintf(stdout,".mSampleRate          = %10.2lf\n",gAudioStreamDescription.mSampleRate);
    fprintf(stdout,".mFormatID            = %d\n",(unsigned int)gAudioStreamDescription.mFormatID);
    //fprintf(stdout,".mFormatID_ByteSwp = %d\n",CFSwapInt32HostToBig(gAudioStreamDescription.mFormatID));
    fprintf(stdout,".mBytesPerPacket      = %d\n",(unsigned int)gAudioStreamDescription.mBytesPerPacket);
    fprintf(stdout,".mFramesPerPacket     = %d\n",(unsigned int)gAudioStreamDescription.mFramesPerPacket);
    fprintf(stdout,".mBytesPerFrame       = %d\n",(unsigned int)gAudioStreamDescription.mBytesPerFrame);
    fprintf(stdout,".mChannelsPerFrame    = %d\n",(unsigned int)gAudioStreamDescription.mChannelsPerFrame);
    //fprintf(stdout,".mBytesPerFrame     = %d\n",(unsigned int)gAudioStreamDescription.mBytesPerFrame);
    fprintf(stdout,".mBitsPerChannel      = %d\n",(unsigned int)gAudioStreamDescription.mBitsPerChannel);
    fprintf(stdout,".mFormatFlags         = %d\n",(unsigned int)gAudioStreamDescription.mFormatFlags);
    fprintf(stdout,"_mFormatHumanReadable = %s\n", descriptionForFormatFlags(DeviceFormat.mFormatFlags));
    
    
    //apply format to output scope //is this really needed?
     DesiredFormat.mFormatID         = gAudioStreamDescription.mFormatID;
    DesiredFormat.mBytesPerPacket   = gAudioStreamDescription.mBytesPerPacket;
    if(DeviceFormat.mFormatFlags == 41)
     DesiredFormat.mFramesPerPacket  = gAudioStreamDescription.mFramesPerPacket;//
    DesiredFormat.mBytesPerFrame    = gAudioStreamDescription.mBytesPerFrame;
     DesiredFormat.mChannelsPerFrame = gAudioStreamDescription.mChannelsPerFrame;
    DesiredFormat.mBytesPerFrame    = gAudioStreamDescription.mBytesPerFrame;
    DesiredFormat.mBitsPerChannel   = gAudioStreamDescription.mBitsPerChannel;
    DesiredFormat.mFormatFlags      = gAudioStreamDescription.mFormatFlags;
    
   
    if(1)lErr = AudioUnitSetProperty(
                         gInputUnitComponentInstance,
                         kAudioUnitProperty_StreamFormat,
                         kAudioUnitScope_Output,
                         kInputBus,
                         &DesiredFormat,
                         sizeof(AudioStreamBasicDescription));
    
    return lErr;
} // vAudioOutputUnitCurrentDeviceSet_04()

// ----------------------------------------------------------------
OSStatus vAudioOutputUnitSetInputCallback_05() {
// ----------------------------------------------------------------
    OSStatus lErr =noErr;

    fprintf(stdout,"------------- (5) Set OutputAU input callback (5) -------------\n");
    AURenderCallbackStruct lInput;
    // force our callback to be recognized as such:
    lInput.inputProc = (AURenderCallback) vAURenderCallback;
    lInput.inputProcRefCon = 0;
    
    lErr = AudioUnitSetProperty(
                         gInputUnitComponentInstance,
                         kAudioOutputUnitProperty_SetInputCallback,
                         kAudioUnitScope_Global,
                         kOutputBus,
                         &lInput,
                         sizeof(lInput));
    return lErr;
} // vAudioOutputUnitSetInputCallback_05()

// ----------------------------------------------------------------
OSStatus vAudioBuffersAllocate_06(UInt32 *step) {
// ----------------------------------------------------------------
    fprintf(stdout,"------------- (6) Allocate internal buffers   (6) -------------\n");    

    OSStatus err = noErr;
    UInt32 expectedInNumberFrames = 0;
    UInt32 propSize = sizeof(UInt32);
    AudioUnitGetProperty(gInputUnitComponentInstance,
                         kAudioDevicePropertyBufferFrameSize,
                         kAudioUnitScope_Global,
                         0,
                         &expectedInNumberFrames,
                         &propSize);
    //expectedInNumberFrames = 256;
    *step = expectedInNumberFrames;
    UInt32 size_in_bytes = (*step)*sizeof(Float32);
    printf("step = %d\n", (unsigned int)(*step));

    
    // allocate AudioBufferList with AudioBuffers:
    UInt32 Bpc = 4;    
    int chans = gAudioStreamDescription.mChannelsPerFrame;
    gAudioBufferList = (AudioBufferList *) malloc (offsetof(AudioBufferList, mBuffers[chans]));
    gAudioBufferList->mNumberBuffers = chans;    
    for (int i = 0; i < chans; i++){
        gAudioBufferList->mBuffers[i].mNumberChannels = 1;
        gAudioBufferList->mBuffers[i].mData = malloc(Bpc*size_in_bytes);
        gAudioBufferList->mBuffers[i].mDataByteSize = Bpc*size_in_bytes;
    }
    
    
    UInt32 flag = 0;
    if(1) //disabling buffer allocation causes bugs on some input devices...
    err = AudioUnitSetProperty(gInputUnitComponentInstance,
                                  kAudioUnitProperty_ShouldAllocateBuffer,
                                  kAudioUnitScope_Input,
                                  kInputBus,
                                  &flag,
                                  sizeof(flag));
    
    return noErr;
} // vAudioBuffersAllocate_06()

// ----------------------------------------------------------------
OSStatus vAudioOutputUnitInitStart_07() {
// ----------------------------------------------------------------
    OSStatus lErr = noErr;
    fprintf(stdout,"------------- (7) Initializing & Starting AU  (7) -------------\n\n");    
    // The AUHAL is now set up to receive input from a device.
    // You must initialize and start the Audio Unit to begin acquiring data.
    lErr = AudioUnitInitialize(gInputUnitComponentInstance);
    if (lErr) return lErr;
    lErr = AudioOutputUnitStart(gInputUnitComponentInstance);
    return lErr;
} // vAudioOutputUnitInitStart_07()
// ----------------------------------------------------------------
void clean_exit(void)
{
    
    AudioOutputUnitStop(gInputUnitComponentInstance);
    AudioUnitUninitialize(gInputUnitComponentInstance);
    AudioComponentInstanceDispose(gInputUnitComponentInstance);
    /*
    free(win); free(bwin); free(bigbuf);   free(lastphase); free(blastphase);
    free(mComplexData);    free(realpart); free(imagpart);
    free(bmComplexData);   free(brealpart);free(bimagpart);
    free(ampData);  free(frqData);
    free(bampData); free(bfrqData);
    for( int i = 0 ; i < 2; i++) free(gStepAudioBuffers[i]);
    free(gStepAudioBuffers);
    */
    printf("Resources released, program cleanly exiting\n");
    exit(0);
}
void info(void)
{
    printf("============== VIEW: version %s =============\n", version);
    printf("program for visualization of mic input signal\n");
    printf("\twritten by: Stanko Juzbasic\n");
    printf("\tusing CoreAudio, vDSP & GLUT API\n");
    printf("\tbased on code by Dmitrij Hmeljak\n");
    printf("\tFor more info type: ?\n\n");    
    
}
// ----------------------------------------------------------------
void usage(void)
{
    printf("\nUSAGE:\n");
    printf("\tview fft_size\n\n");
    printf("\tavailable fft_sizes: 1024, 2048, 4096, 8192\n\n");
    printf("\tinteractive controls: b, d, f, w, l, p, s, =, -\n");
    printf("\t                     'up', 'dn', '^up', '^dn'\n\n");
    printf("\tb    : bands resolution full display\n");
    printf("\ts    : sonogram display (only with bands on)\n");
    printf("\td    : decibell display\n");
    printf("\tf    : exact frequency display\n");
    printf("\tw    : windowed input (default : on)\n");
    printf("\tl    : log. frequency  display\n");
    printf("\tp    : polygon/line display\n");
    printf("\t=    : increase graphic spectral resolution\n");
    printf("\t-    : decrease graphic spectral resolution\n");
    printf("\t'up' : \n");
    printf("\t'dn' : translate spectral graph (in full & exact mode)\n");
    printf("\t'^up': \n");
    printf("\t'^dn': reset spectral graph position\n\n");
    
}

// ----------------------------------------------------------------
# pragma mark - main -

// ----------------------------------------------------------------
int main (int argc, const char * argv[]) {
// ----------------------------------------------------------------
    
    //_______________(0) Parse command line arguments______________
    OSStatus err = noErr;
    int mult = 4;
    if (argc > 2){usage(); exit(-1);}
    
    if(argc == 2 ){
    if(strncmp(argv[1], "-h",  2) == 0){usage(); exit(-1);}
    if(strncmp(argv[1], "-?",  2) == 0){usage(); exit(-1);}
    if(strncmp(argv[1], "--h", 3) == 0){usage(); exit(-1);}
    if(strncmp(argv[1], "?"  , 1) == 0){usage(); exit(-1);}
    
    if (atoi(argv[1]) == !atoi(argv[1])){usage(); exit(-1);}//NaN check
    if (atoi(argv[1]) <= 1024){mult = 2; goto one;}
    if (atoi(argv[1]) <= 2048){mult = 4; goto one;}
    if (atoi(argv[1]) <= 4096){mult = 8; goto one;}
    if (atoi(argv[1]) >  4096){mult = 16; goto one;}
    }
    
    // ------------- (1) Open an AudioOutputUnit (1) -----------------
    // You must first obtain an AudioOutputUnit by using a Component Description,
    // as you would when attempting to obtain any other Audio Unit.
    one:
    err = vAUOutputOpen_01();
    // ------------- (2) Enable AUHAL's IO for input (2) -------------
    err = vAUHALEnableInputIO_02();
    // ------------- (3) Set default input device current input of AUHAL (3) -------------
    err = vAUHALSetDefaultInputDev_03();
    // ------------- (3) Set default input device current input of AUHAL (3) -------------
    err = vGetIOBufferrange_03_1();
    // ------------- (4) Set the  current device of the AudioOutputUnit  (4) -------------
    err = vAudioOutputUnitCurrentDeviceSet_04();
    // ------------- (5) Set the Input callback for the AudioOutputUnit  (5) -------------
    err = vAudioOutputUnitSetInputCallback_05();

    // ------------- (6) Allocating necessary buffers (6) -------------
    err = vAudioBuffersAllocate_06 (&D);
    if(D != AUHAL_SAMPLES) mult = D/AUHAL_SAMPLES;
    N = mult*D;
    
    size_t sz = (size_t)R;
    win  = (float*)calloc(sz, sizeof(float));
    bwin = (float*)calloc(sz, sizeof(float));
    cheby_win( win, D, 100);
    cheby_win(bwin, N, 100);
    windowing = true;
    mFFTSetup  = vDSP_create_fftsetup ( log2(D), 2 );
    bmFFTSetup = vDSP_create_fftsetup ( log2(N), 2 );
    fundamental  = R/(float)D;
    bfundamental = R/(float)N;
    gY = 2;
    printf("D  = %d\n", (unsigned int)D);
    printf("N  = %d\n", (unsigned int)N);
    N2 = N/2;
    DN = D*N;
    D1 = 64;
    D1N= D1*N;
    count = D1N - N2;
    printf("D1N = %u\n", D1N);
    megamp  = (Float32*)calloc(D1N+1, sizeof(Float32)); megamphead  = megamp + count;
    megfreq = (Float32*)calloc(D1N+1, sizeof(Float32)); megfreqhead = megfreq+ count;
    //AUHAL_SAMPLES = D;
        
    // ------------- (7) Initializing and Starting the AudioOutputUnit (7) -------------

    
    gNpointsToDisplay =  AUHAL_SAMPLES;
    poly = TRUE;
        
    printf("============= VIEW: version %s ===========\n", version);
    {
        struct timespec sleepinterval;
        int n;
        sleepinterval.tv_sec  = 0;
        sleepinterval.tv_nsec = 100000000;
        n = nanosleep (&sleepinterval,NULL);  /* sleep according to POSIX */
    }        
    printf("============= HAL on, starting GL ==========\n");
    GlOutputStart();
    
    err = vAudioOutputUnitInitStart_07();
    

    fprintf(stderr, "%s", "starting glutMainLoop... \n");
    
    glutMainLoop();

    return 0;
} // main


// -----------------------------------
// initializes OpenGL and GLUT
// -----------------------------------
static void GlOutputStart() {
    //this is buggy and has to have exactly this format and order
    int foo = 1;
    char * bar[1] = {" "};
    
    char text[256];
    strcpy(text, "initializing OpenGL...\n");

    fprintf(stdout,"%s",text);

    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_ALPHA);
    glutInitWindowPosition( MY02TEST_WIND_X_POS, MY02TEST_WIND_Y_POS );
    glutInitWindowSize( MY02TEST_WIND_X_SIZE, MY02TEST_WIND_Y_SIZE );
    
    // GLUT initialization here
    glutInit(&foo, bar);
    
    // creating window for OpenGL output:
    //    wion't show until glutMainLoop() is called.
    glutCreateWindow("input signal view");
    printf("window created\n");
    
    glEnable( GL_BLEND );
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    /// GLUT Event Handlers:
    glutDisplayFunc       (GlDisplay);
    glutReshapeFunc       (reshape);
    glutIdleFunc          (GlIdle); 
    glutKeyboardFunc      (GlKey);
    glutSpecialFunc       (GlSpecial);
    
    GlInit();            // initialize OpenGL projection
    
    
} // GlOutputStart()



// ------------------------------------------------------------------------
static void GlInit (void) {
    int width_viewport, hight_viewport;   //  to store the size of the window
    
    
    //  get the current window size from GLUT
    width_viewport = MAX(glutGet(GLUT_WINDOW_WIDTH) , MY02TEST_WIND_X_MINSIZE);
    hight_viewport = MAX(glutGet(GLUT_WINDOW_HEIGHT) , MY02TEST_WIND_Y_MINSIZE);
    if(0)    printf("glut_w_W=%d, glut_w_H=%d \n", width_viewport, hight_viewport);
    
    // this time, force window at least 200x200
    glutReshapeWindow(width_viewport, hight_viewport);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    // use the GL projection matrix: we want to initialize that one
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // gloOrtho2D(left,right,bottom,top) creates a 2D projection matrix
    //    where lower-left corner  is at x,y=(left,bottom)
    //    and upper-right corner is at x,y=(right,top) 
    /// this also takes care of the GLUT vs. OpenGL coordinate mismatch: 
    gluOrtho2D(0.0,
               width_viewport,
               0.0,
               hight_viewport);
    // call glViewport() to clip the display to the new window size 
    glViewport(0, 0, width_viewport, hight_viewport);
    
    
    glPointSize(1.0);
    glLineWidth(1.0);
    
} // GlInit()



// ------------------------------------------------------------------------
static void GlDisplay(void) {
    int i;
    //int j;
    int lWHeight, lWWidth, lBarWidth;
    static float lColorF = 0.0f;
    static float prevhl, prevhr;
    
    if(b && sonogram && !logfrq){
    memcpy(megamphead, bampData, N2*sizeof(Float32));
    memcpy(megfreqhead,bfrqData, N2*sizeof(Float32));
    count -= N2;
    
    if(count > 0){
     megfreqhead -= N2;
     megamphead  -= N2;
    }else{
     count        = D1N - N2;
     megfreqhead  = megfreq + count;
     megamphead   = megamp  + count;
    }
    //printf ("count: %u\n", count);
    }
    
    lColorF = lColorF+0.0001f;if (lColorF>1.0f)lColorF = 0.0001f;
    if(1) glClearColor( 0., 0., 0.2f, 1.0f );
    else glClearColor( lColorF, lColorF, 0.3f, 1.0f );
    glClear(GL_COLOR_BUFFER_BIT);  // clear  the color buffer!

    lWWidth = glutGet(GLUT_WINDOW_WIDTH);
    lWHeight = glutGet(GLUT_WINDOW_HEIGHT);
    lBarWidth = MIN(1, (int)((float)lWWidth/(float)(AUHAL_SAMPLES+2)));
    
    // now **** here **** start the real drawing of window contens
    float fact = lWWidth/(float)MY02TEST_WIND_X_SIZE;
    float w    = 10.*fact;
    float vact = lWHeight/(float)MY02TEST_WIND_Y_SIZE;    
    
    float zeroL = (float)(lWHeight/4. +10.0f );
    float zeroR = (float)(3*lWHeight/4. +10.0f );
    float lamplF, lamprF, logamplF, logamprF;

    for (i=0; i<gNpointsToDisplay; i++) {
        float lColor2F = ((float)i) / ((float)gNpointsToDisplay);

        lamplF = gvAudioBuffers[0][i]*gn*100.0f;
        if(gvAudioBuffers[0][i] > 0.)
        logamplF = lgn*amp2dbf(gvAudioBuffers[0][i], minvalue);
        else
        logamplF = 1.*lgn*amp2dbf(fabsf(gvAudioBuffers[0][i]), minvalue);
        
        glColor3f( 1.0f, 0.0f, lColor2F );        
        if(poly){
        glBegin(GL_POLYGON);
          glVertex2f( w+(float)i*fact,    zeroL );
          glVertex2f( w+(float)(i+1)*fact,zeroL );
          glVertex2f( w+(float)(i+1)*fact,zeroL+1.+ (dB ? logamplF:lamplF ));
          glVertex2f( w+(float)i*fact,    zeroL+1.+prevhl);
        glEnd();
        }else{
        glBegin(GL_LINE_STRIP);
          glVertex2f( w+(float)i*fact,       zeroL+prevhl);
          glVertex2f( w+(float)(i+1.)*fact,  zeroL+(dB ? logamplF:lamplF ));
        glEnd();
        }
        
        prevhl = (dB ? logamplF:lamplF);

        lamprF = gvAudioBuffers[1][i]*gn*100.0f;
        if(gvAudioBuffers[1][i] > 0.)          
        logamprF = lgn*amp2dbf(gvAudioBuffers[1][i], minvalue);
        else
        logamprF = 1.*lgn*amp2dbf(fabsf(gvAudioBuffers[1][i]), minvalue);
        
        glColor3f( 0.0f, 1.0f, lColor2F );        
        if(poly){
        glBegin(GL_POLYGON);
          glVertex2f( w+(float)i*fact,     zeroR );
          glVertex2f( w+(float)(i+1)*fact, zeroR );
          glVertex2f( w+(float)(i+1)*fact, zeroR+1.+(dB ? logamprF:lamprF ) );
          glVertex2f( w+(float)i*fact,     zeroR+1.+prevhr );
        glEnd();
        }else{
       glBegin(GL_LINE_STRIP);
        glVertex2f( w+(float)i*fact,       zeroR+prevhr );
        glVertex2f( w+(float)(i+1)*fact,   zeroR+(dB ? logamprF:lamprF ) );
       glEnd();
        }
        if(i == 0)if(0)printf("%f\t%f\n", prevhr, lamprF);
        prevhr = (dB ? logamprF: lamprF);
    }//end for i=0...
    
    //do spectral drawing-------------------------------
    float minfrq;
    float logmaxfreq = logf((float)(R*.5));
    float xpos = w+gNpointsToDisplay*fact;

    float g = (windowing ? 0.8 : 0.3);
    float r = (poly ? 0.7 : 1.0);
    float bl = (freqDisp ? .4 : .1);
    glColor4f(r,g,bl,0.5);
    //float maxp = 0.;
    Float32 p, prevp, fy, prevfy;
    if(b)prevp = bampData[0];
    else prevp =  ampData[0];
    float don = D/(float)N;
    int imax;
    if(b){imax = N/2; minfrq= R/N;}
    else {imax = D/2; minfrq= R/D;}
    
    if(!b && gY < 2)gY = 2;
    
    float v = (( b&&freqDisp ) ? vt : 0. );
    
    if(freqDisp){        
        if(b){
            if(logfrq){
              prevfy = 1.;  
            }
            else prevfy = don*(float)gY*bfrqData[0]/(bfundamental);
        } else {
            if(logfrq){
                prevfy = 1.;
            }else prevfy = (float)gY*frqData[0]/(fundamental);
        }        
    } else {
     if(logfrq)  prevfy = 1.;
     else prevfy = 0.;
    }
    
    
for(i = 1; i < imax; i++){        
    if(freqDisp){
        if(b){
            if(logfrq){
                  fy     = D*logf(2*D*bfrqData[i]/(N*bfundamental))/(.69314718*logmaxfreq);
                //printf("%f\n",bfrqData[i]/(i*bfundamental ));
            }else fy     = don*(float)gY*bfrqData[i]/(bfundamental);
        }else{
            if(logfrq){
                  fy     = D*logf(2*frqData[i]/fundamental)/(.69314718*logmaxfreq);
                //printf("fy = %f\n",fy);
            }else fy     = (float)gY*frqData[i]/(fundamental);
        }
    } else {
        if(logfrq){
          //if(b)fy = D*logf(2*(i))/(.69314718*logmaxfreq);
          //else
              fy = D*logf(2*(i))/(.69314718*logmaxfreq);
        }
        else fy = (float)gY*i;//gain in Y direction, default = 2;        
    }
            
    if(b) p = bampData[i];
    else  p = ampData [i];
        
        if(poly){
        if(dB){
            Float32 dbpp= amp2dbf( prevp, .000005);//.05
            Float32 dbp = amp2dbf( p,     .000005);
            
            glBegin(GL_POLYGON);
            glVertex2f(xpos+1  , 40. + prevfy*vact + v);
            glVertex2f(xpos+1  , 40. + fy*vact + v);
            glVertex2f(xpos  - 5.*dbp*fact, 40.  + fy*vact + v);
            glVertex2f(xpos  - 5.*dbpp*fact, 40. + prevfy*vact + v);
            glEnd();
        }else{
        glBegin(GL_POLYGON);
        glVertex2f(xpos+1 , 40. + prevfy*vact + v);
        glVertex2f(xpos+1 , 40. + fy*vact + v);
        glVertex2f(xpos - 10000.*p, 40. + fy*vact + v);
        glVertex2f(xpos - 10000.*prevp, 40. + prevfy*vact + v);
        glEnd();
        }
    
        }else{
            if(dB){
                Float32 dbpp= amp2dbf( prevp, .000005);//.05
                Float32 dbp = amp2dbf( p,     .000005);
                
                glBegin(GL_LINE_STRIP);
                glVertex2f(xpos  - 5.*dbpp*fact, 40. + prevfy*vact + v);
                glVertex2f(xpos  - 5.*dbp*fact,  40. + fy*vact + v);
                glEnd();
            }else{
        glBegin(GL_LINE_STRIP);
        glVertex2f(xpos - 10000.*prevp, 40. + prevfy*vact + v);
        glVertex2f(xpos - 10000.*p, 40. + fy*vact + v);
        glEnd();
            }
        }
     prevp = p;
     prevfy = fy;
        //printf("%f\t%d\n",fy*vact , lWHeight );
        if(b)
        if (fy*vact + v + 80. > (float)lWHeight)break;
    } // end for i ...
    
    
    //do sonogram drawing------------------------------------------
    unsigned int count1 = D1N - N2;
    int h = N2 - lWHeight;
//    printf ("count: %u\n", count);
    if(b && sonogram && !logfrq){
        float vs = vact*gY;
        
    for( i = count; i < count1+count; i ++){
        if(i%N2 > lWHeight ) i += h;
        float a = 64.*megamp[i%count1];
        //float a0= 64.*megamp[(i-N2)%count1];
        if (i == count)glColor4f(1., 0., a*.2, 1.);
        //else if(i ==count1-N2)glColor4f(0., 0., 1., 1.);
        // else
        glColor4f((a > .6 ? a : .6), (a > .4 ? a :.4), a*.2, a);
        int x1 =  ((i-count)/N2)*fact;
        //fy =    gY*(i%N2)
        //40. + fy*vact + v = 40. + gY*(i%N2)*vact + v = 40. + vs*(i%N2) +v;
        int y1 =  vs*(i%N2) + v;

    if(surface){
        int x0 =  (i == count ? x1 :((i-N2-count)/N2)*fact);
        int y2 =  vs*((i-1)%N2) + v;
     glBegin(GL_POLYGON);
        glVertex2f(xpos+x0, 40 + y1);
        glVertex2f(xpos+x1, 40 + y1);
        glVertex2f(xpos+x1, 40 + y2);
        glVertex2f(xpos+x0, 40 + y2);
     glEnd();
    }else{
        glPointSize(gY+2);
        glBegin(GL_POINTS);
        glVertex2f(xpos+x1, 40 + y1);
        glEnd();
        }
    }//end for( i = count; ...
}
    glutSwapBuffers();
} // GlDisplay()


// ------------------------------------------------------------------------
static void reshape(int width, int height) {
    GlInit();              //reset OpenGL projection mode
    glutPostRedisplay();   // now tell GLUT it's time to update the window!
} // reshape()

// ------------------------------------------------------------------------
static void GlKey(unsigned char key, int x, int y) {
    
    switch (key) {
        case 27:
        case 'q':
        case 'Q': {
            //exit(0);
            clean_exit();
            }
            break;
        case 'B':
        case 'b':
            b = !b;
            break;
        case 'D':
        case 'd':
            dB = !dB;
            //if (dB) minvalue = 0.00001;
            break;
        case 'F':
        case 'f':
            freqDisp = !freqDisp;
            break;
        case 'i':
        case 'I':
            info();
            break;
        case 'L':
        case 'l':
            //if(freqDisp)
                logfrq = !logfrq;
            break;
        case 'm':
            if(dB){
            minvalue /= 10.;
            if (minvalue < 0.0000001)minvalue = 0.0000001;    
            }
            break;
        case 'M':
            if(dB){
            minvalue *= 10.;
            if (minvalue > 0.01)minvalue = 0.01;
            }
            break;
        case 'P':
        case 'p':
            poly = !poly;
            break;
        case 'R':
        case 'r':
            //surface = !surface;
            break;
        case 'S':
        case 's':
            sonogram = !sonogram;
            break;
            
        case 'W':
        case 'w':            
            windowing = !windowing;
            break;
        case '=':
            gY++;
            if(gY > 20)gY = 20.;
            break;
        case '-':
            gY--;
            if(gY < 1)gY = 1;
            break;
        case '?':
            usage();
            break;
        default:
            break;
    }
} // GlKey()
// ------------------------------------------------------------------------
void GlSpecial(int key, int px, int py)
{
    int i = 0;
    int modifiers = glutGetModifiers();
    gLastKey = key;
    if(modifiers == i){;} //just to silence the compiler
    switch (key) {
        case GLUT_KEY_RIGHT:
            if(dB){
            lgn += .1;
            if(lgn > 4.)lgn = 4.;
            if(0)printf("lgn = %f\n",lgn);
            }else{
            gn += 1.;
            if (gn > 10.) gn = 10.;
            }            
        break;
        case GLUT_KEY_LEFT:
            if(dB){
            lgn -= .1;
            if(lgn < 1.)lgn = 1.;
            if(0)printf("lgn = %f\n",lgn);
            }else{
            gn -= 1.;
            if (gn <= 1.)gn = 1.;
            }            
        break;
        case GLUT_KEY_UP:
            if(freqDisp  && b){
            if(modifiers == GLUT_ACTIVE_CTRL) vt = 0.;
            else vt += 20.;
            }
            break;
        case GLUT_KEY_DOWN:
            if(freqDisp  && b){
            if(modifiers == GLUT_ACTIVE_CTRL) vt = 0.;    
            vt -= 20.;
            }            
            break;
            
        default:
        break;
    }
}//GlSpecial()
// ------------------------------------------------------------------------
static void GlIdle(void) {

    {
        struct timespec sleepinterval;
        int n;
        sleepinterval.tv_sec  = 0;
        sleepinterval.tv_nsec = 50000000; // 50M nSec = 0.05 seconds
        n = nanosleep (&sleepinterval,NULL);  /* sleep according to POSIX */
    }        
    
    glutPostRedisplay();
} // GlIdle()
// -----------------------------------------------------------------------
// -----------------------------------------------------------------------
static void GlBitmapString(int x, int y, char *string)
{
    int len, i;
            
    glRasterPos2f(x, y);
    len = (int) strlen(string);
    for (i = 0; i < len; i++) {
        glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, string[i]);
    }
} // GlBitmapString()

